/*
 * Copyright (c) 2008-2009 NVIDIA Corporation
 *
 * Based on sample OpenGL code bearing this notice:
 *
 * This code was created by Jeff Molofee '99 
 * (ported to Linux/GLX by Mihael Vrbanec '00)
 *
 * If you've found this code useful, please let me know.
 *
 * Visit Jeff at http://nehe.gamedev.net/
 * 
 * or for port-specific comments, questions, bugreports etc. 
 * email to Mihael.Vrbanec@stud.uni-karlsruhe.de
 */
 
#include <dlfcn.h>
#include <stdio.h>
#include <math.h>
#define GL_GLEXT_PROTOTYPES
#include <GL/glx.h>
#include <GL/gl.h>
#include <GL/glext.h>
#include <GL/glu.h>
#include <vdpau/vdpau_x11.h>
#include "traceplay.h"
#include "common.h"
#include "common_gl.h"

VdpDevice           vdp_device;
VdpGetProcAddress * vdp_get_proc_address;
VdpDeviceDestroy *  vdp_device_destroy;

PFNGLVDPAUINITNVPROC                  ptrglVDPAUInitNV;
PFNGLVDPAUFININVPROC                  ptrglVDPAUFiniNV;
PFNGLVDPAUREGISTEROUTPUTSURFACENVPROC ptrglVDPAURegisterOutputSurfaceNV;
PFNGLVDPAUUNREGISTERSURFACENVPROC     ptrglVDPAUUnregisterSurfaceNV;
PFNGLVDPAUMAPSURFACESNVPROC           ptrglVDPAUMapSurfacesNV;
PFNGLVDPAUUNMAPSURFACESNVPROC         ptrglVDPAUUnmapSurfacesNV;

/* stuff about our window grouped together */
typedef struct {
    Display *dpy;
    int screen;
    Window win;
    GLXContext ctx;
    XSetWindowAttributes attr;
    int x, y;
    unsigned int width, height;
    unsigned int depth;    
} GLWindow;

typedef struct {
    GLvdpauSurfaceNV glSurface;
    GLuint           glTexName;
} RegisteredSurface;

/* attributes for a single buffered visual in RGBA format with at least
 * 4 bits per color and a 16 bit depth buffer */
static int attrListSgl[] = {
    GLX_RGBA, GLX_RED_SIZE, 4, 
    GLX_GREEN_SIZE, 4, 
    GLX_BLUE_SIZE, 4, 
    GLX_DEPTH_SIZE, 16,
    None
};

/* attributes for a double buffered visual in RGBA format with at least
 * 8 bits per color and a 16 bit depth buffer */
static int attrListDbl[] = {
    GLX_RGBA, GLX_DOUBLEBUFFER, 
    GLX_RED_SIZE, 8, 
    GLX_GREEN_SIZE, 8, 
    GLX_BLUE_SIZE, 8, 
    GLX_DEPTH_SIZE, 16,
    None
};

GLWindow GLWin;
GLfloat rotX = 0.6;  /* X Rotation */
GLfloat rotY = 0.8;  /* Y Rotation */
GLfloat rotZ = 1.2;  /* Z Rotation */

int               rsi;
RegisteredSurface rs[8];

/* function called when our window is resized (should only happen in window mode) */
void resizeGLScene(unsigned int width, unsigned int height)
{
    if (height == 0)    /* Prevent A Divide By Zero If The Window Is Too Small */
        height = 1;
    /* Reset The Current Viewport And Perspective Transformation */
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f, (GLfloat)width / (GLfloat)height, 0.1f, 100.0f);
    glMatrixMode(GL_MODELVIEW);

}

/* general OpenGL initialization function */
void initGL(GLvoid)
{
    glEnable(GL_TEXTURE_RECTANGLE_ARB);
    glShadeModel(GL_SMOOTH);
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    /* we use resizeGLScene once to set up our initial perspective */
    resizeGLScene(GLWin.width, GLWin.height);
    glFlush();
}

void * traceplay_video_surface_post_create(
    struct Traceplay_Ctx * ctx,
    VdpVideoSurface        surface
)
{
    return 0;
}

void * traceplay_output_surface_post_create(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface
)
{
    glGenTextures(1, &rs[rsi].glTexName);
    rs[rsi].glSurface = ptrglVDPAURegisterOutputSurfaceNV((GLvoid *)(size_t)surface, GL_TEXTURE_RECTANGLE_ARB, 1, &rs[rsi].glTexName);
    void * surface_context = (void *)(size_t)rsi;
    rsi++;
    return surface_context;
}

void traceplay_video_surface_pre_destroy(
    struct Traceplay_Ctx * ctx,
    VdpVideoSurface        surface,
    void *                 surface_context
)
{
}

void traceplay_output_surface_pre_destroy(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
    ptrglVDPAUUnregisterSurfaceNV(rs[(size_t)surface_context].glSurface);
}

void traceplay_display_video_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpVideoSurface        surface,
    void *                 surface_context
)
{
    printf("ERROR at %s:%d: display_video_surface called\n", __FILE__, __LINE__);
    common_exit_error();
}

void traceplay_wait_for_output_surface_idle(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
}

void drawGLScene(uint32_t frame_number, GLuint tex);

void traceplay_display_output_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
    ptrglVDPAUMapSurfacesNV(1, &rs[(size_t)surface_context].glSurface);

    if (opt_gold_surface_type == SURFACE_TYPE_GL_INTEROP_OUT) {
        common_gl_do_gold_gl_tex_rgb(frame_number, GL_TEXTURE_RECTANGLE_ARB, rs[(size_t)surface_context].glTexName);
    }

    drawGLScene(frame_number, rs[(size_t)surface_context].glTexName);

    ptrglVDPAUUnmapSurfacesNV(1, &rs[(size_t)surface_context].glSurface);
}

void traceplay_process_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpVideoSurface        vid_surface,
    void *                 vid_surface_context,
    VdpOutputSurface       out_surface,
    void *                 out_surface_context
)
{
    printf("ERROR at %s:%d: process_surface called\n", __FILE__, __LINE__);
    common_exit_error();
}

Traceplay_Callbacks tpcb = {
    traceplay_video_surface_post_create,
    traceplay_output_surface_post_create,
    traceplay_video_surface_pre_destroy,
    traceplay_output_surface_pre_destroy,
    traceplay_display_video_surface,
    traceplay_wait_for_output_surface_idle,
    traceplay_display_output_surface,
    traceplay_process_surface
};

void initVDPAU(void)
{
    VdpStatus vdp_st;

    // Create Device

    vdp_st = vdp_device_create_x11(
        GLWin.dpy,
        GLWin.screen,
        &vdp_device,
        &vdp_get_proc_address
    );
    CHECK_ST

    // Entry point retrieval

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_DEVICE_DESTROY,
        (void **)&vdp_device_destroy
    );
    CHECK_ST

    void * dlh = dlopen(NULL, RTLD_LAZY);
    if (!dlh) {
        printf("ERROR at %s:%d: dlopen(NULL) failed: %s\n", __FILE__, __LINE__, dlerror());
        common_exit_error();
    }

    ptrglVDPAUInitNV = (PFNGLVDPAUINITNVPROC)(uintptr_t)dlsym(dlh, "glVDPAUInitNV");
    if (!ptrglVDPAUInitNV) {
        printf("ERROR at %s:%d: dlsym() failed: %s\n", __FILE__, __LINE__, dlerror());
        common_exit_error();
    }

    ptrglVDPAUFiniNV = (PFNGLVDPAUFININVPROC)(uintptr_t)dlsym(dlh, "glVDPAUFiniNV");
    if (!ptrglVDPAUFiniNV) {
        printf("ERROR at %s:%d: dlsym() failed: %s\n", __FILE__, __LINE__, dlerror());
        common_exit_error();
    }

    ptrglVDPAURegisterOutputSurfaceNV = (PFNGLVDPAUREGISTEROUTPUTSURFACENVPROC)(uintptr_t)dlsym(dlh, "glVDPAURegisterOutputSurfaceNV");
    if (!ptrglVDPAURegisterOutputSurfaceNV) {
        printf("ERROR at %s:%d: dlsym() failed: %s\n", __FILE__, __LINE__, dlerror());
        common_exit_error();
    }

    ptrglVDPAUUnregisterSurfaceNV = (PFNGLVDPAUUNREGISTERSURFACENVPROC)(uintptr_t)dlsym(dlh, "glVDPAUUnregisterSurfaceNV");
    if (!ptrglVDPAUUnregisterSurfaceNV) {
        printf("ERROR at %s:%d: dlsym() failed: %s\n", __FILE__, __LINE__, dlerror());
        common_exit_error();
    }

    ptrglVDPAUMapSurfacesNV = (PFNGLVDPAUMAPSURFACESNVPROC)(uintptr_t)dlsym(dlh, "glVDPAUMapSurfacesNV");
    if (!ptrglVDPAUMapSurfacesNV) {
        printf("ERROR at %s:%d: dlsym() failed: %s\n", __FILE__, __LINE__, dlerror());
        common_exit_error();
    }

    ptrglVDPAUUnmapSurfacesNV = (PFNGLVDPAUUNMAPSURFACESNVPROC)(uintptr_t)dlsym(dlh, "glVDPAUUnmapSurfacesNV");
    if (!ptrglVDPAUUnmapSurfacesNV) {
        printf("ERROR at %s:%d: dlsym() failed: %s\n", __FILE__, __LINE__, dlerror());
        common_exit_error();
    }

    ptrglVDPAUInitNV((void *)(size_t)vdp_device, (void *)(uintptr_t)vdp_get_proc_address);
}

void finiVDPAU(void)
{
    ptrglVDPAUFiniNV();

    vdp_device_destroy(
        vdp_device
    );
}

/* Here goes our drawing code */
void drawGLScene(uint32_t frame_number, GLuint tex)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
    glTranslatef(0.0f, 0.0f, -5.0f);
    glRotatef(rotX, 1.0f, 0.0f, 0.0f);  /* rotate on the X axis */
    glRotatef(rotY, 0.0f, 1.0f, 0.0f);  /* rotate on the Y axis */
    glRotatef(rotZ, 0.0f, 0.0f, 1.0f);  /* rotate on the Z axis */

    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, tex);
    glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

#define glTexCoord2f(x, y) glTexCoord2f((x) * trace_stream_conf_ptr->out_width, (1.0 - (y)) * trace_stream_conf_ptr->out_height);

    glBegin(GL_QUADS);
        /* front face */
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(-1.0f, -1.0f, 1.0f); 
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(1.0f, -1.0f, 1.0f);
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(1.0f, 1.0f, 1.0f);
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(-1.0f, 1.0f, 1.0f);
        /* back face */
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(-1.0f, -1.0f, -1.0f); 
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(-1.0f, 1.0f, -1.0f);
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(1.0f, 1.0f, -1.0f);
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(1.0f, -1.0f, -1.0f);
        /* right face */
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(1.0f, -1.0f, -1.0f); 
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(1.0f, 1.0f, -1.0f);
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(1.0f, 1.0f, 1.0f);
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(1.0f, -1.0f, 1.0f);
        /* left face */
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(-1.0f, -1.0f, 1.0f); 
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(-1.0f, 1.0f, 1.0f);
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(-1.0f, 1.0f, -1.0f);
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(-1.0f, -1.0f, -1.0f);
        /* top face */
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(1.0f, 1.0f, 1.0f); 
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(1.0f, 1.0f, -1.0f);
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(-1.0f, 1.0f, -1.0f);
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(-1.0f, 1.0f, 1.0f);
        /* bottom face */
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(1.0f, -1.0f, -1.0f); 
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(1.0f, -1.0f, 1.0f);
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(-1.0f, -1.0f, 1.0f);
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(-1.0f, -1.0f, -1.0f);
    glEnd();

    if (opt_gold_surface_type == SURFACE_TYPE_GL_WIN) {
        common_gl_do_gold_window(frame_number);
    }

    glXSwapBuffers(GLWin.dpy, GLWin.win);

    /* change the rotation angles */
    rotX += 0.3f;
    rotY += 0.2f;
    rotZ += 0.4f;
}

/* function to release/destroy our resources and restoring the old desktop */
GLvoid killGLWindow(GLvoid)
{
    if (GLWin.ctx) {
        if (!glXMakeCurrent(GLWin.dpy, None, NULL)) {
            printf("Could not release drawing context.\n");
        }
        glXDestroyContext(GLWin.dpy, GLWin.ctx);
        GLWin.ctx = NULL;
    }
    XCloseDisplay(GLWin.dpy);
}

/* this function creates our window and sets it up properly */
/* FIXME: bits is currently unused */
void createGLWindow(char* title, int width, int height, int bits)
{
    XVisualInfo *vi;
    Colormap cmap;
    int glxMajorVersion, glxMinorVersion;
    Atom wmDelete;
    Window winDummy;
    unsigned int borderDummy;
    
    /* get a connection */
    GLWin.dpy = XOpenDisplay(0);
    if (!GLWin.dpy) {
        printf("ERROR at %s:%d: XOpenDisplay failed\n", __FILE__, __LINE__);
        common_exit_error();
    }
    GLWin.screen = DefaultScreen(GLWin.dpy);
    /* get an appropriate visual */
    vi = glXChooseVisual(GLWin.dpy, GLWin.screen, attrListDbl);
    if (vi == NULL) {
        vi = glXChooseVisual(GLWin.dpy, GLWin.screen, attrListSgl);
        printf("Only Singlebuffered Visual!\n");
    }
    else {
        printf("Got Doublebuffered Visual!\n");
    }
    glXQueryVersion(GLWin.dpy, &glxMajorVersion, &glxMinorVersion);
    printf("glX-Version %d.%d\n", glxMajorVersion, glxMinorVersion);
    /* create a GLX context */
    GLWin.ctx = glXCreateContext(GLWin.dpy, vi, 0, GL_TRUE);
    /* create a color map */
    cmap = XCreateColormap(GLWin.dpy, RootWindow(GLWin.dpy, vi->screen),
        vi->visual, AllocNone);
    GLWin.attr.colormap = cmap;
    GLWin.attr.border_pixel = 0;

    /* create a window in window mode*/
    GLWin.attr.event_mask = ExposureMask | KeyPressMask | ButtonPressMask |
        StructureNotifyMask;
    GLWin.win = XCreateWindow(GLWin.dpy, RootWindow(GLWin.dpy, vi->screen),
        0, 0, width, height, 0, vi->depth, InputOutput, vi->visual,
        CWBorderPixel | CWColormap | CWEventMask, &GLWin.attr);
    /* only set window title and handle wm_delete_events if in windowed mode */
    wmDelete = XInternAtom(GLWin.dpy, "WM_DELETE_WINDOW", True);
    XSetWMProtocols(GLWin.dpy, GLWin.win, &wmDelete, 1);
    XSetStandardProperties(GLWin.dpy, GLWin.win, title,
        title, None, NULL, 0, NULL);
    XMapRaised(GLWin.dpy, GLWin.win);

    /* connect the glx-context to the window */
    glXMakeCurrent(GLWin.dpy, GLWin.win, GLWin.ctx);
    XGetGeometry(GLWin.dpy, GLWin.win, &winDummy, &GLWin.x, &GLWin.y,
        &GLWin.width, &GLWin.height, &borderDummy, &GLWin.depth);
    printf("Depth %d\n", GLWin.depth);
    if (glXIsDirect(GLWin.dpy, GLWin.ctx)) 
        printf("Congrats, you have Direct Rendering!\n");
    else
        printf("Sorry, no Direct Rendering possible!\n");
    initGL();
}

void help()
{
    printf("app [options]\n");
    common_print_help();
}

int main(int argc, char * * argv)
{
    int ret;

    app_name = "traceplay_gl_out";
    support_opt_gold_surface_type_mask =
        (1 << SURFACE_TYPE_VDP_VID_DEC_ORDER) |
        (1 << SURFACE_TYPE_VDP_VID_DISP_ORDER) |
        (1 << SURFACE_TYPE_VDP_OUT) |
        (1 << SURFACE_TYPE_GL_INTEROP_OUT) |
        (1 << SURFACE_TYPE_GL_WIN);

    common_init();

    argc--;
    argv++;
    while (argc) {
        if ((ret = common_parse_arg(&argc, &argv)) != COMMON_PARSE_ARG_NOT_PARSED) {
            if (ret != COMMON_PARSE_ARG_OK) {
                help();
                common_exit_error();
            }
        }
        else {
            printf("ERROR: Unknown option '%s'\n", argv[0]);
            help();
            common_exit_error();
        }
        argc--;
        argv++;
    }

    if (common_check_args() != COMMON_CHECK_ARGS_OK) {
        help();
        common_exit_error();
    }

    common_load_stream();

#define WIN_WIDTH  1280
#define WIN_HEIGHT 1024
    createGLWindow(
        "VDPAU/GL Interop (VdpOutputSurface)",
        WIN_WIDTH,
        WIN_HEIGHT,
        24
    );
    initVDPAU();
    common_gl_playback_init(
        vdp_device,
        vdp_get_proc_address,
        WIN_WIDTH,
        WIN_HEIGHT
    );
    struct Traceplay_Ctx *traceplay_ctx = traceplay_init(
        vdp_device,
        vdp_get_proc_address,
        TRACEPLAY_PLAY_TYPE_OUTPUT,
        &tpcb,
        trace_stream_conf_ptr
    );
    CHECK_PTR(traceplay_ctx)
    traceplay_play(
        traceplay_ctx,
        trace_cmds_ptr,
        opt_repeat_count
    );
    traceplay_fini(
        traceplay_ctx
    );
    common_gl_playback_fini();
    finiVDPAU();
    killGLWindow();

    common_exit_ok();
    return 0;
}

